#include "lsqrms.h"
#include "umath.h"
#include "provepdb.h"
#include "matvecalloc.h"
#include "dynapro.h"
#include "matvecutils.h"

/* cRMS = 0.44*e^(2.22*[fraction mutated residues]) -
   Koehl and Levitt, JMB 323(3):551-562, 2002 */
#define L_MIN 0.44		/* fraction mutated = 0 */
#define L_MIN_SQR (L_MIN*L_MIN)
#define L_THRESH 0.96		/* fraction mutated = 0.35 */
#define L_THRESH_SQR (L_THRESH*L_THRESH)
#define L_THRESH2 2.08		/* fraction mutated = 0.7 */
#define L_THRESH2_SQR (L_THRESH2*L_THRESH2)
#define BAD_THRESH 3.626 /* fraction mutated = 0.95 */

#define epsilon2 MAX(epsilon,1e-10)

static void usage(int status) {
  fprintf(stderr,
	  "lsqrms [-sGhao1] [-b|-B|-i|-n|-g|-p] [-l|-L|-t] source.pdb target.pdb [structal.out]\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -s: generate STRUCTAL-like output\n");
  fprintf(stderr, "  -l: use LSQ_ONLY fit instead of iterated RMSD\n");
  fprintf(stderr, "  -L: scale <= %f to 0 (%f thresh)\n", L_MIN, L_THRESH);
  fprintf(stderr, "  -t: scale <= %f to 0 (%f thresh)\n", L_MIN, L_THRESH2);
  fprintf(stderr, "  -G: don't use gap penalties for sequence alignment\n");
  fprintf(stderr, "  -h: print this usage message\n");
  fprintf(stderr, "  -1: OK if single-residue 'alignment' (also does -o)\n");
  fprintf(stderr, "  -b: use blosum62 for initial alignment\n");
  fprintf(stderr, "  -B: use positive-only blosum62 for initial alignment\n");
  fprintf(stderr, "  -n: use Nussinov matrix for initial alignment\n");
  fprintf(stderr, "  -i: use identity matrix for initial alignment\n");
  fprintf(stderr, "  -g: use gonnet_p matrix for initial alignment\n");
  fprintf(stderr, "  -p: use pam120_p matrix for initial alignment\n");
  fprintf(stderr,
	  "  -a: use altered residues (__[_0-9]) - cannot be used with -i\n");
  fprintf(stderr, "  -o: use O, C, N as well as CA (also done by -1)\n");
  exit(status);
}

int main(int argc, char* argv[]) {
  double RMS, RMSold, PEQUIV, tempval, RMSsav, quality, quality_old;
  double RMSmin = 1000000;
  double quality_max = 0;
  double quality_sav = 0;
  int i, j;
  int rows, cols;
  double ** ss;
  double ** SS;
  double ** dd;
  int * iOpnPen;
  int total_OpnPen = 0;
  double * iExtPen;
  int * jOpnPen;
  double * jExtPen;
  double ** TrRow;
  double ** TrCol;
  int iOverallMax;
  int jOverallMax;

  int opt_c;
  int opt_errflag = 0;
  extern int optind;
  
  int *iina, *iina_sav, *iina_old;
  int *jinb, *jinb_sav, *jinb_old;
  int nbrks, nbrks_old, nEqPairs, nbrks_sav;
  
  float fp1, fp2, fp3;
  char   ** aa_que, ** aa_tar;
  double ** x_que, **x_que_original;
  double ** x_tar;
  double ** x_que_sav, ** x_que_old;
  double * Tcm;
  FILE *stream;
  char *char_stream, *char_stream_backup, *char_stream_head;
  char *structal_out;
  extern int errno;
  
  char *s2, *s3, *a_que, *a_tar;
  int d2;
  int *r_num_que, *r_num_tar;
  int StringMax = 4;
  
  rotate_t U;
  xlate_t  T;
  
  int stay, nEqPairs_old, nEqPairs_sav;
  int iter = 0;
  int again = 0;
  int again2 = 0;
  int sav_updated = 0;
  int increased_gap_penalties = 0;
  char aa;
  
  double *TMPvec, *Tzero, /* *Tcm_que, */ *Tcm_tar; 
  double **Rtemp, **Runity;
  double angle_b, angle_c/*, angle_b_que, angle_c_que */;
  double angle_b_tar, angle_c_tar;
  double xdir_que, /* ydir_que, */ rdir_que;
  double xdir_tar, /* ydir_tar, */ rdir_tar;
  int iMaxDist;
  
  CoordType *str1, *str2;

  /* Flag to use OLD output similar to the one from STRUCTAL*/
  int structal     = 0;

  /* Flag to use LSQ_ONLY fit instead of iterational RMSD */
  int lsq          = 0;

  /* Modified by Julian Graham, April 2003 */

  /* Then by Allen Smith, 2003-2007 */

  int use_blosum62 = 0;
  int use_blosum62_pos = 0;
  int use_nussinov = 0;
  int use_identity = 0;
  int use_gonnet_p = 0;
  int use_pam120_p = 0;
  int use_gap = 1;
  int altered = 0;
  int use_carbonyl = 0;
  int use_min_thresh = 0;
  int use_min_thresh2 = 0;
  int single_ok = 0;

  while ((opt_c = getopt(argc, argv, "slLbBiGhaongtp1")) != -1) {
    switch (opt_c) {
    case 'a':
      if (use_identity)
	opt_errflag = 1;
      else
	altered = 1;
      break;

    case 'o':
      use_carbonyl = 1;
      break;

    case 's':
      structal = 1;
      break;
      
    case '1':
      single_ok = 1;
      use_carbonyl = 1;
      break;

    case 'l':
      if (use_min_thresh || use_min_thresh2)
	opt_errflag = 1;
      else
	lsq = 1;
      break;

    case 'L':
      if (lsq || use_min_thresh2)
	opt_errflag = 1;
      else
	use_min_thresh = 1;
      break;

    case 't':
      if (lsq || use_min_thresh)
	opt_errflag = 1;
      else
	use_min_thresh2 = 1;
      break;

    case 'p':
      if (use_identity || use_nussinov || use_gonnet_p || use_blosum62 ||
	  use_blosum62_pos)
	opt_errflag = 1;
      else
	use_pam120_p = 1;
      break;

    case 'b':
      if (use_identity || use_nussinov || use_gonnet_p || use_pam120_p)
	opt_errflag = 1;
      else
	use_blosum62 = 1;
      break;

    case 'B':
      if (use_identity || use_nussinov || use_gonnet_p || use_pam120_p)
	opt_errflag = 1;
      else
	use_blosum62_pos = 1;
      break;

    case 'i':
      if (use_blosum62 || altered || use_nussinov || use_pam120_p ||
	  use_blosum62_pos || use_gonnet_p)
	opt_errflag = 1;
      else
	use_identity = 1;
      break;

    case 'n':
      if (use_blosum62 || use_identity || use_blosum62_pos ||
	  use_gonnet_p || use_pam120_p)
	opt_errflag = 1;
      else
	use_nussinov = 1;
      break;

    case 'g':
      if (use_blosum62 || use_identity || use_blosum62_pos ||
	  use_nussinov || use_pam120_p)
	opt_errflag = 1;
      else
	use_gonnet_p = 1;
      break;

    case 'G':
      use_gap = 0;
      break;

    case 'h':
      usage(0);
      /* NOTREACHED */
      break;

    case '?':
      opt_errflag = 1;
      break;
    }
  }

  if (opt_errflag)
    usage(2);

  if ((argc - optind) < 1) {
    fprintf (stderr, "LSQRMS: need to specify (2) .PDB filenames!\n");
    usage(2);
  }
  
  printf("Program LSQRMS. Version 2.02-Allen-10/26/2005\n");
  if (structal) {
    if ((argc - optind) < 2)
      structal_out = strdup("structal.out");
    else
      structal_out = strdup(argv[optind+2]);
    printf("Using Structal output format (output to %s)\n", structal_out);
  }

  if (lsq)
    printf("Doing LSQ_ONLY fit instead of iterated RMSD\n");
  else if (use_min_thresh)
    printf("Scaling <= %f to 0 (%f threshold)\n", L_MIN, L_THRESH);
  else if (use_min_thresh2)
    printf("Scaling <= %f to 0 (%f threshold)\n", L_MIN, L_THRESH2);

  if (altered)
    printf("Using altered residues for site/whatever alignment\n");
  if (use_carbonyl)
    printf("Using O, C, N as well as CA\n");

  if (use_blosum62_pos)
    printf("Using positive-only blosum62 for initial alignment\n");
  else if (use_blosum62)
    printf("Using blosum62 for initial alignment\n");
  else if (use_nussinov)
    printf("Using Nussinov (Naor et al 1996) matrix for initial alignment\n");
  else if (use_identity)
    printf("Using identity matrix for initial alignment\n");
  else if (use_gonnet_p)
    printf("Using gonnet_p matrix for initial alignment\n");
  else if (use_pam120_p)
    printf("Using pam120_p matrix for initial alignment\n");

  if (use_gap)
    printf("Using gap penalties for initial alignment\n");
  else
    printf("Not using gap penalties for initial alignment\n");
  fflush(NULL);

  stay = 1; nEqPairs_sav = -1; nEqPairs = -1; nEqPairs_old = -1;
  
  /* s1 = cVector(1,StringMax); */
  s2 = cVector(1,StringMax);
  s3 = cVector(1,3);
  /* s4 = cVector(1,StringMax); */
  
  /* ATOM      2  CA  ASP H  19      30.313   5.590  12.619  */
  /* ATOM      3  O   ASP H  19      30.313   5.590  12.619  */
  char_stream = provepdb(argv[optind],altered,use_carbonyl);
  if (char_stream == NULL) {
    fprintf (stderr, "Couldn't open file \'%s\'\n", argv[optind]);
    return 1; 
  } else {
    char_stream_backup = strdup(char_stream);
    if (char_stream_backup == NULL) {
      fprintf(stderr, "Out of memory for strdup of char_stream for \'%s\'\n",
	      argv[optind]);
      return 1;
    }
    char_stream_head = char_stream;
    i=0;
    while (1) {
      char *new_pointer;

      (void) sscanf(char_stream,
		    "ATOM %*d %2c %3[^ ] %d %f %f %f\n", 
		    s2, s3, &d2, &fp1, &fp2, &fp3);
      new_pointer = strstr(char_stream + 4, "ATOM");
      i++;
      if (new_pointer == NULL)
	break;
      else {
	char_stream = (char *) (new_pointer);
	continue;
      }
    } /* while (1) */
  }
  rows = i;

  free(char_stream_head);
  char_stream = char_stream_backup;
  char_stream_head = char_stream;

  x_que     = dMatrix(1,3,0,rows-1);
  x_que_sav = dMatrix(1,3,0,rows-1);
  x_que_old = dMatrix(1,3,0,rows-1);
  x_que_original = dMatrix(1,3,0,rows-1);
  a_que     = cVector(0,rows-1);
  aa_que    = cMatrix(0,rows-1,0,2);
  r_num_que = iVector(0,rows-1);

  for (i = 0; i < rows; i++) {
    char *new_pointer = strstr (char_stream + 4, "ATOM");

    if (sscanf(char_stream,
	       "ATOM %*d %2c %3[^ ] %d %f %f %f\n", 
	       s2, s3, &d2, &fp1, &fp2, &fp3) == 6) {
      char_stream = (char *) (new_pointer);
      aa = convert321(s3,1,altered);
      if (use_carbonyl && isalpha(aa)) {
	if (islower(aa))
	  aa = _toupper(aa);
	if (s2[0] == 'C') {
	  if (s2[1] != 'A')
	    aa = tohigh(_tolower(aa));
	} else if (s2[0] == 'O')
	  aa = _tolower(aa);
	else
	  aa = tohigh(aa);
      }
      a_que[i] = aa;
      r_num_que[i] = d2;
      aa_que[i][0] = s3[0];
      aa_que[i][1] = s3[1];
      aa_que[i][2] = s3[2];
      
      x_que[1][i] = fp1;
      x_que_sav[1][i] = fp1;
      x_que_old[1][i] = fp1;
      x_que_original[1][i] = fp1;
      
      x_que[2][i] = fp2;
      x_que_sav[2][i] = fp2;
      x_que_old[2][i] = fp2;
      x_que_original[2][i] = fp2;
      
      x_que[3][i] = fp3;
      x_que_sav[3][i] = fp3;
      x_que_old[3][i] = fp3;
      x_que_original[3][i] = fp3;
    } else {
      fprintf(stderr, "Unable to convert %s\n", char_stream);
      return 1;
    }
  } /* for(i) */
  
  free(char_stream_head);
  char_stream = provepdb(argv[optind+1],altered,use_carbonyl);
  if (char_stream == NULL) {
    fprintf (stderr, "Couldn't open file \'%s\'\n", argv[optind+1]);
    return 1;
  } else {
    char_stream_backup = strdup (char_stream);
    if (char_stream_backup == NULL) {
      fprintf(stderr, "Out of memory for strdup of char_stream for \'%s\'\n",
	      argv[optind+1]);
      return 1;
    }
    char_stream_head = char_stream;
    i=0;
    while (1) {
      char *new_pointer;

      (void) sscanf(char_stream,
		    "ATOM %*d %2c %3[^ ] %d %f %f %f\n", 
		    s2, s3, &d2, &fp1, &fp2, &fp3);
      new_pointer = strstr (char_stream + 4, "ATOM");
      i++;
      if (new_pointer == NULL)
	break;
      else {
	char_stream = (char *) (new_pointer);
	continue;
      }
    } /* while (1) */
  }
  cols = i;
  
  free (char_stream_head);
  char_stream = char_stream_backup;
  char_stream_head = char_stream;
  /* End of modification by Julian Graham, April 2003 */
  
  x_tar     = dMatrix(1,3,0,cols-1);
  a_tar     = cVector(0,cols-1);
  aa_tar    = cMatrix(0,cols-1,0,2);
  r_num_tar = iVector(0,cols-1);
  
  for (i = 0; i < cols; i++) {
    char *new_pointer = strstr (char_stream + 4, "ATOM");

    if (sscanf(char_stream,
	       "ATOM %*d %2c %3[^ ] %d %f %f %f\n", 
	       s2, s3, &d2, &fp1, &fp2, &fp3) == 6) {
      char_stream = (char *) (new_pointer);
      aa = convert321(s3,1,altered);
      if (use_carbonyl && isalpha(aa)) {
	if (islower(aa))
	  aa = _toupper(aa);
	if (s2[0] == 'C') {
	  if (s2[1] != 'A')
	    aa = tohigh(_tolower(aa));
	} else if (s2[0] == 'O')
	  aa = _tolower(aa);
	else
	  aa = tohigh(aa);
      }
      a_tar[i] = aa;
      r_num_tar[i] = d2;
      aa_tar[i][0] = s3[0];
      aa_tar[i][1] = s3[1];
      aa_tar[i][2] = s3[2];
      x_tar[1][i] = fp1; 
      x_tar[2][i] = fp2; 
      x_tar[3][i] = fp3;
    } else {
      fprintf(stderr, "Unable (2) to interpret %s\n",char_stream);
      return 1;
    }
  } /* for (i < cols) */

  free(char_stream_head);
  /* ATOM     10  CA  GLU H  20      30.998   8.719  10.360  */
  
  iOverallMax = -1;
  jOverallMax = -1;
  
  ss      = dMatrix(0,rows+1,0,cols+1);
  SS      = dMatrix(0,rows+1,0,cols+1);
  dd      = dMatrix(0,rows+1,0,cols+1);
  TrRow   = dMatrix(0,rows,0,cols);
  TrCol   = dMatrix(0,rows,0,cols);
  iOpnPen = iVector(0,rows-1);
  iExtPen = dVector(0,rows-1);
  jOpnPen = iVector(0,cols-1);
  jExtPen = dVector(0,cols-1);
  
  iina_sav = iVector(0,MAXSEQ);
  jinb_sav = iVector(0,MAXSEQ);
  
  
  /* FIRST OF ALL, MOVE que->CM(tar); NEXT, align PRINCIPAL AXES OF INERTIA: */
  
  Tcm     = dVector(1,3); 
  /* Tcm_que = dVector(1,3); */
  Tcm_tar = dVector(1,3); 
  Tzero   = dVector(1,3); 
  TMPvec  = dVector(1,3); 
  Rtemp   = dMatrix(1,3,1,3);
  Runity  = dMatrix(1,3,1,3);

  Rtemp[1][1]  = 1.0;
  Rtemp[1][2]  = 0.0;
  Rtemp[1][3]  = 0.0;

  Rtemp[2][1]  = 0.0;
  Rtemp[2][2]  = 1.0;
  Rtemp[2][3]  = 0.0;

  Rtemp[3][1]  = 0.0;
  Rtemp[3][2]  = 0.0;
  Rtemp[3][3]  = 1.0;

  Runity[1][1] = 1.0;
  Runity[1][2] = 0.0;
  Runity[1][3] = 0.0;

  Runity[2][1] = 0.0;
  Runity[2][2] = 1.0;
  Runity[2][3] = 0.0;

  Runity[3][1] = 0.0;
  Runity[3][2] = 0.0;
  Runity[3][3] = 1.0;

  TMPvec[1] = 2;
  TMPvec[2] = 3;
  TMPvec[3] = 4;

  Tzero[1] = 00.e0;
  Tzero[2] = 0.0e0;
  Tzero[3] = 0.0e0;
 
  (void)findCM(x_tar,cols,Tcm);

  /* saved inv translation */
  Tcm_tar[1] = Tcm[1];
  Tcm_tar[2] = Tcm[2];
  Tcm_tar[3] = Tcm[3];

  Tcm[1] = -Tcm[1];
  Tcm[2] = -Tcm[2];
  Tcm[3] = -Tcm[3];

  (void)TRmoveXYZstruc(Tcm,Runity,x_tar,cols);
  
  (void)findCM(x_tar,cols,Tcm);
  
  iMaxDist = FindMaxXYZDistIdxLast(x_tar,cols);
  (void)AlignZ(x_tar[1][iMaxDist],x_tar[2][iMaxDist],x_tar[3][iMaxDist],
	       &angle_b,&angle_c);
  /* saved inv rotation */
  angle_b_tar = -angle_b;
  angle_c_tar = -angle_c;

  (void)FormRotMatrY(Rtemp,angle_b);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);
  (void)FormRotMatrX(Rtemp,angle_c);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);

  iMaxDist = FindMaxXYDistIdxLast(x_tar,cols);

  rdir_tar = sqrt( SQR(x_tar[1][iMaxDist]) + SQR(x_tar[2][iMaxDist]) );
  /* xdir_tar: needed to determine XY-plane rotation */
  xdir_tar = x_tar[1][iMaxDist]/rdir_tar;
  /* ydir_tar = x_tar[2][iMaxDist]/rdir_tar; */

  /* Do the same for the QUERY: */
  (void)findCM(x_que,rows,Tcm);
  /* saved inv translation */
  /* Tcm_que[1] = Tcm[1];
     Tcm_que[2] = Tcm[2];
     Tcm_que[3] = Tcm[3]; */

  Tcm[1] = -Tcm[1];
  Tcm[2] = -Tcm[2];
  Tcm[3] = -Tcm[3];

  (void)TRmoveXYZstruc(Tcm,Runity,x_que,rows);
  printf("Moving query: %.10f %.10f %.10f\n", Tcm[1], Tcm[2], Tcm[3]);
  
  (void)findCM(x_que,rows,Tcm);
  
  iMaxDist = FindMaxXYZDistIdxLast(x_que,rows);
  (void)AlignZ(x_que[1][iMaxDist],x_que[2][iMaxDist],x_que[3][iMaxDist],
	       &angle_b,&angle_c);
  /* saved inv rotation - not needed (Allen) */
  /* angle_b_que = -angle_b; angle_c_que = -angle_c; */

  (void)FormRotMatrY(Rtemp,angle_b);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);
  printf("Rotating query: y-angle %.10f; %.10f %.1f %.10f; %.1f %.1f %.1f; %.10f %.1f %.10f\n",
	 angle_b,
	 Rtemp[1][1],Rtemp[1][2],Rtemp[1][3],
	 Rtemp[2][1],Rtemp[2][2],Rtemp[2][3],
	 Rtemp[3][1],Rtemp[3][2],Rtemp[3][3]);
  (void)FormRotMatrX(Rtemp,angle_c);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);
  printf("Rotating query: x-angle %.10f; %.1f %.1f %.1f; %.1f %.10f %.10f; %.1f %.10f %.10f\n",
	 angle_c,
	 Rtemp[1][1],Rtemp[1][2],Rtemp[1][3],
	 Rtemp[2][1],Rtemp[2][2],Rtemp[2][3],
	 Rtemp[3][1],Rtemp[3][2],Rtemp[3][3]);

  iMaxDist = FindMaxXYDistIdxLast(x_que,rows);

  rdir_que = sqrt( SQR(x_que[1][iMaxDist]) + SQR(x_que[2][iMaxDist]) );
  /* xdir_que: needed to determine XY-plane rotation */
  xdir_que = x_que[1][iMaxDist]/rdir_que;
  /* ydir_que = x_que[2][iMaxDist]/rdir_que; */

  angle_b = acos(xdir_tar);
  angle_c = acos(xdir_que);

  (void)FormRotMatrZ(Rtemp,angle_c-angle_b);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);
  printf("Rotating query: z-angle %.10f; %.10f %.10f %.1f; %.10f %.10f %.1f; %.1f %.1f %.1f\n",
	 (angle_c-angle_b),
	 Rtemp[1][1],Rtemp[1][2],Rtemp[1][3],
	 Rtemp[2][1],Rtemp[2][2],Rtemp[2][3],
	 Rtemp[3][1],Rtemp[3][2],Rtemp[3][3]);

  /* CHECK COORDINATES: */
/*   stream = fopen( "coordchk.txt", "w" ); */
/*   if( stream == NULL ) */
/*     printf( "Couldn't open file coordchk.txt\n" ); */
/*   else { */
/*     fprintf( stream, "Query: \n"); */
/*     for (i=0;i<rows;i++) { */
/*       fprintf( stream, "%12.3f %8.3f %8.3f\n", */
/* 	       x_que[1][i], x_que[2][i], x_que[3][i]); */
/*     } */
/*     fprintf( stream, "Target: \n"); */
/*     for (i=0;i<cols;i++) { */
/*       fprintf( stream, "%12.3f %8.3f %8.3f\n", */
/* 	       x_tar[1][i], x_tar[2][i], x_tar[3][i]); */
/*     } */
/*     fclose( stream ); */
/*   } */
 
  /* INITIAL STARTING POINT: SEQUENCE ALIGNMENT.
     Decide if needed (based on PID) */
  iina = iVector(0,MAXSEQ);
  jinb = iVector(0,MAXSEQ);
  iina_old = iVector(0,MAXSEQ);
  jinb_old = iVector(0,MAXSEQ);
  for (i=0; i<=MAXSEQ; i++) {
    iina[i] = -1;
    jinb[i] = -1;
  }
  for (j=0; j<=(cols-1); j++) {
    if (use_gap && (j < (cols-1))) {
/* values from J. Mol. Biol. (1995) 249, 816-831:
   An Assessment of Amino Acid Exchange Matrices in
   Aligning Protein Sequences: The Twilight Zone
   Revisited, by Gerhard Vogt, Thure Etzold and
   Patrick Argos */
      /* NOTE: values use Opn + ((len-1)*Ext) formula */
      if (use_nussinov) { /* estimated from pam60_p values */
	jOpnPen[j] = 14;
	jExtPen[j] = 1.5;
      } else if (use_gonnet_p) {
	jOpnPen[j] = 6;
	jExtPen[j] = 0.8;
      } else if (use_blosum62_pos) {
	jOpnPen[j] = 20;
	jExtPen[j] = 1.2;
      } else if (use_blosum62) {
	jOpnPen[j] = 15;
	jExtPen[j] = 1.8;
      } else if (use_identity) {
	jOpnPen[j] = 1;
	jExtPen[j] = 0.07;
      } else if (use_pam120_p) {
	jOpnPen[j] = 12.5;
	jExtPen[j] = 1;
      } else {
	jOpnPen[j] = 15;
	jExtPen[j] = 0.92;
      }
      if ((cols > rows) && (j >= (rows-1)))
	jOpnPen[j] = 0;
      if ((toascii(a_tar[j]) == 'X') || (toascii(a_tar[j]) == 'x'))
	jExtPen[j] = 0.0;
    } else {
      jOpnPen[j] = 0;
      jExtPen[j] = 0.0;
    }
  }
  for (i=0; i<=(rows-1); i++) {
    if (use_gap && (i < (rows-1))) {
      if (use_nussinov) { /* estimated from pam60_p values */
	iOpnPen[i] = 14;
	iExtPen[i] = 1.5;
      } else if (use_gonnet_p) {
	iOpnPen[i] = 6;
	iExtPen[i] = 0.8;
      } else if (use_blosum62_pos) {
	iOpnPen[i] = 20;
	iExtPen[i] = 1.2;
      } else if (use_blosum62) {
	iOpnPen[i] = 15;
	iExtPen[i] = 1.8;
      } else if (use_identity) {
	iOpnPen[i] = 1;
	iExtPen[i] = 0.07;
      } else if (use_pam120_p) {
	iOpnPen[i] = 12.5;
	iExtPen[i] = 1;
      } else {
	iOpnPen[i] = 15;
	iExtPen[i] = 0.92;
      }
      if ((rows > cols) && (i >= (cols-1)))
	iOpnPen[i] = 0;
      if ((toascii(a_que[i]) == 'X') || (toascii(a_que[i]) == 'x'))
	iExtPen[i] = 0.0;
    } else {
      iOpnPen[i] = 0;
      iExtPen[i] = 0.0;
    }
    for (j=0; j<=(cols-1); j++) {       
      SS[i][j]    = 0.0;
      TrRow[i][j] = 0.0;
      TrCol[i][j] = 0.0;
    }
  }
  /* Use substitution matrix over here: */
  for (i=0; i<rows; i++) {
    for(j=0; j<cols; j++) {
      if (use_identity || use_blosum62 || use_blosum62_pos) {
	if (a_que[i] == a_tar[j])
	  ss[i][j] = 1.0; /* primitive! */
	else
	  ss[i][j] = 0.0; /* primitive! */
	if (use_blosum62 || use_blosum62_pos) {
	  /* blosum62 enhancement, not always good */
	  tempval = blosum62(a_que[i],a_tar[j],altered,use_carbonyl);
	  ss[i][j] += tempval;
	  if (use_blosum62_pos)
	    ss[i][j] += 8;
	} else if (use_carbonyl) {
	  if (isascii(a_que[i]) != isascii(a_tar[j]))
	    ss[i][j] = -1000;
	  else if (! isascii(a_que[i])) {
	    if (isupper(toascii(a_que[i])) != isupper(toascii(a_tar[j])))
	      ss[i][j] = -1000;
	    else if (isalpha(toascii(a_que[i])) !=
		     isalpha(toascii(a_tar[j])))
	      ss[i][j] = -1000;
	  } else if (isupper(a_que[i]) != isupper(a_tar[j]))
	    ss[i][j] = -1000;
	}
      } else if (use_nussinov) {
	ss[i][j] = 0.0;
	tempval = nussinov(a_que[i],a_tar[j],altered,use_carbonyl);
	ss[i][j] += tempval;
      } else if (use_gonnet_p) {
	ss[i][j] = 0.0;
	tempval = gonnet_p(a_que[i],a_tar[j],altered,use_carbonyl);
	ss[i][j] += tempval;
      } else if (use_pam120_p) {
	ss[i][j] = 0.0;
	tempval = pam120_p(a_que[i],a_tar[j],altered,use_carbonyl);
	ss[i][j] += tempval;
      } else {
	ss[i][j] = 0.0;
	tempval = glyproaln(a_que[i],a_tar[j],altered,use_carbonyl);
	ss[i][j] += tempval;
      }
    }
  }

  printf("Have %d rows, %d cols\n", rows, cols);
  fflush(NULL);

  (void)dyn_pr_n2(ss, iOpnPen, iExtPen, jOpnPen, jExtPen, r_num_que, r_num_tar,
		  rows, cols, SS, TrRow, TrCol, &iOverallMax, &jOverallMax); 
  nbrks = -1;
  nEqPairs_sav = nEqPairs;
  nEqPairs_old = nEqPairs;
  nEqPairs = TrMat2Equivalences(rows,cols,TrRow,TrCol,
				iOverallMax,jOverallMax,
				iina,jinb,&nbrks);
  printf("nEqPairs: %d\n", nEqPairs); fflush(NULL);
  nbrks_sav = nbrks;
  nbrks_old = nbrks;
  for (i=0; i<=MAXSEQ; i++) {
    iina_sav[i] = iina[i];
    iina_old[i] = iina[i];
    jinb_sav[i] = jinb[i];
    jinb_old[i] = jinb[i];
  }

  tempval = MIN((double)rows, (double)cols);
  PEQUIV = (double)nEqPairs/(tempval);

  if ((single_ok && (nEqPairs >= tempval) && (nEqPairs >= 4)) ||
      ((nEqPairs > 2) && (lsq || (PEQUIV >= 0.3)) &&
       ((nEqPairs > 8) || (! use_carbonyl)))) {
    /* It makes sense to start with eq.pairs from sequence alignment: */

    if (use_carbonyl)
      nEqPairs = 4*(floor((double)nEqPairs/4.0));

    str1 = (CoordType *)malloc(nEqPairs * sizeof(CoordType));
    if (!str1) {
      fprintf(stderr, "main: Allocation failed!\n");
      exit(1);
    }
    str2 = (CoordType *)malloc(nEqPairs * sizeof(CoordType));
    if (!str2) {
      fprintf(stderr, "main: Allocation failed!\n");
      exit(1);
    }

    for (i=0;i<nEqPairs;i++) {
      str1[i].x = x_que[1][iina[i]-1];
      str2[i].x = x_tar[1][jinb[i]-1];
      
      str1[i].y = x_que[2][iina[i]-1];
      str2[i].y = x_tar[2][jinb[i]-1];
      
      str1[i].z = x_que[3][iina[i]-1];
      str2[i].z = x_tar[3][jinb[i]-1];
    }

    RMS = bestfit(U, T, str1, str2, nEqPairs);
    if (RMS > 0.01)
      quality = fsqrt(nEqPairs)/RMS;
    else
      quality = 100*fsqrt(nEqPairs);
    printf("PEQUIV, RMS, eq.pairs, quality (from seq align) = %f %f %d %f\n",
	   PEQUIV*100,RMS,nEqPairs,quality);
    quality_max = quality;
    quality_sav = quality;
    quality_old = quality;
    RMSsav = RMS;
    RMSold = RMS;
    for(i=0; i<rows; i++) {
      x_que_sav[1][i] = x_que[1][i];
      x_que_sav[2][i] = x_que[2][i];
      x_que_sav[3][i] = x_que[3][i];
      x_que_old[1][i] = x_que[1][i];
      x_que_old[2][i] = x_que[2][i];
      x_que_old[3][i] = x_que[3][i];
    }

    (void)rotrans(x_que, rows, U, T);
    
    if (RMS < RMSmin)
      RMSmin = RMS;
    free((RELEASE)(str1));
    free((RELEASE)(str2));
  } else { /* PID < threshold */
    printf("PEQUIV is %f, rather low to start with %d SeqAli Eq. pairs\n",
	   PEQUIV*100, nEqPairs);
  }

  free_iVector(iina,0, MAXSEQ);
  free_iVector(jinb,0, MAXSEQ);

  fflush(NULL);

  /* END OF THE INITIAL STARTING POINT */
  
  if (!lsq) {
    for (j=0; j<=(cols-1); j++) {
      if (altered && isascii(a_tar[j]) && (! isalpha(a_tar[j]))) {
	jOpnPen[j] = 20;
	jExtPen[j] = 20;
      } else {
	jOpnPen[j] = 0;
	jExtPen[j] = 0;
      }
    }
    for (i=0; i<=(rows-1); i++) {
      if (altered && isascii(a_que[i]) && (! isalpha(a_que[i]))) {
	iOpnPen[i] = 20;
	iExtPen[i] = 20;
      } else {
	iOpnPen[i] = 0;
	iExtPen[i] = 0;
      }
    }

    do {
      iter++;    /* START OF MAIN ITERATION CYCLE: */
      /* CHECK COORDINATES: */
      /*      stream = fopen( "coordchk.dat", "a" ); */
      /*      if (stream == NULL) */
      /*        printf("Couldn't open file coordchk.dat\n" ); */
      /*      else { */
      /*        fprintf(stream, "Iteration # %d: \n",iter); */
      /*        fprintf(stream, "Query: \n"); */
      /*        for (i=0; i<rows; i++) { */
      /* 	 fprintf(stream, "%f %f %f\n", */
      /* 		 x_que[1][i], x_que[2][i], x_que[3][i]); */
      /*        } */
      /*        fprintf(stream, "Target: \n"); */
      /*        for (i=0; i<cols; i++) { */
      /* 	 fprintf(stream, "%f %f %f\n", */
      /* 		 x_tar[1][i], x_tar[2][i], x_tar[3][i]); */
      /*        } */
      /*        fclose(stream); */
      /*      } */

      again = 0;
      again2 = 0;
      sav_updated = 0;
      increased_gap_penalties = 0;
     
      iina = iVector(0,MAXSEQ);
      jinb = iVector(0,MAXSEQ);

      for (i=0; i<=MAXSEQ; i++) {
	iina[i] = -1;
	jinb[i] = -1;
      }

      for (i=0; i<=rows-1; i++) {
	for (j=0;j<=cols-1;j++) {
	  dd[i][j] = (SQR(x_que[1][i]-x_tar[1][j]) +
			SQR(x_que[2][i]-x_tar[2][j]) +
			SQR(x_que[3][i]-x_tar[3][j]));
	  ss[i][j] = 0;
	  if (use_carbonyl) {
	    tempval = check_carbonyl(a_que[i],a_tar[j]);
	    ss[i][j] += tempval;
	  }
	  if (ss[i][j] == 0) {
	    ss[i][j] = dd[i][j];
	    if (use_min_thresh && (ss[i][j] < L_THRESH_SQR)) {
	      if (ss[i][j] <= L_MIN_SQR)
		ss[i][j] = 0;
	      else {
		ss[i][j] =
		  (sqrt(ss[i][j]) - L_MIN)*(L_THRESH/(L_THRESH-L_MIN));
		if (ss[i][j] > 0) /* in case of roundoff error */
		  ss[i][j] *= ss[i][j];
		else
		  ss[i][j] = 0;
	      }
	    } else if (use_min_thresh2 && (ss[i][j] < L_THRESH2_SQR)) {
	      if (ss[i][j] <= L_MIN_SQR)
		ss[i][j] = 0;
	      else {
		ss[i][j] =
		  (sqrt(ss[i][j]) - L_MIN)*(L_THRESH2/(L_THRESH2-L_MIN));
		if (ss[i][j] > 0) /* in case of roundoff error */
		  ss[i][j] *= ss[i][j];
		else
		  ss[i][j] = 0;
	      }
	    }
	    ss[i][j] = 20.0/(1.0 + (ss[i][j]/25.0));
	    dd[i][j] = sqrt(dd[i][j]);
	    
	    /* ss[i][j] =
	       sqrt((x_que[1][i]-x_tar[1][j])*(x_que[1][i]-x_tar[1][j]) +
	       (x_que[2][i]-x_tar[2][j])*(x_que[2][i]-x_tar[2][j]) +
	       (x_que[3][i]-x_tar[3][j])*(x_que[3][i]-x_tar[3][j]));
	       
		 ss[i][j] = 20.0 / ( 1.0 + SQR(ss[i][j]/5.0) ); */
	  }

	  SS[i][j] = 0.0;
	  TrRow[i][j] = 0.0;
	  TrCol[i][j] = 0.0;
	}
      }

      (void)dyn_pr_n2(ss, iOpnPen, iExtPen, jOpnPen, jExtPen,
		      r_num_que, r_num_tar, rows, cols,
		      SS,  TrRow,   TrCol, &iOverallMax, &jOverallMax); 

      nbrks_old = nbrks;
      nbrks = -1;

      nEqPairs_old = nEqPairs;
      quality_old = quality;
      nEqPairs = TrMat2Equivalences(rows,cols,TrRow,TrCol,
				    iOverallMax,jOverallMax,
				    iina,jinb,&nbrks);

      if (use_carbonyl)
	nEqPairs = 4*(floor((double)nEqPairs/4.0));
      
      str1 = (CoordType *)malloc(nEqPairs * sizeof(CoordType));
      if (!str1) {
	fprintf(stderr,"main: Allocation failed!\n");
	exit(1);
      }
      str2 = (CoordType *)malloc(nEqPairs * sizeof(CoordType));
      if (!str2) {
	fprintf(stderr,"main: Allocation failed!\n");
	exit(1);
      }

      for (i=0; i<nEqPairs; i++) {
	str1[i].x = x_que[1][iina[i]-1];
	str2[i].x = x_tar[1][jinb[i]-1];

	str1[i].y = x_que[2][iina[i]-1];
	str2[i].y = x_tar[2][jinb[i]-1];

	str1[i].z = x_que[3][iina[i]-1];
	str2[i].z = x_tar[3][jinb[i]-1];
      }

      RMSold = RMS;
      RMS = bestfit(U, T, str1, str2, nEqPairs);
      if (RMS > 0.01)
	quality = fsqrt(nEqPairs)/RMS;
      else
	quality = 100*fsqrt(nEqPairs);
      printf("RMS, eq.pairs, quality = %f %d %f\n",RMS,nEqPairs,quality);

      (void)rotrans(x_que, rows, U, T);

      /* expected increase in RMS is proportional to square root of chain
	 length - Allen */

      if ((RMS < RMSmin) || 
	  ((quality > quality_sav) && ((RMS < BAD_THRESH) ||
				       (RMS == RMSmin))) ||
	  ((nEqPairs_sav < 0) &&
	   ((nEqPairs > nEqPairs_old) || (quality >= quality_sav) ||
	    (RMS <= RMSsav) ||
	    ((nEqPairs < nEqPairs_old) && (quality < quality_sav) &&
	     (total_OpnPen < 10)))) ||
	  ((nbrks < nbrks_sav) && (quality_sav <= quality) &&
	   ((RMS < BAD_THRESH) || (RMS <= RMSsav))) ||
	  ((nEqPairs_sav < nEqPairs) &&
	   ((RMS < BAD_THRESH) || (RMS <= RMSmin)) &&
	   ((RMS - RMSsav) < 0.1))) {
	if ((nEqPairs_sav < 0) && (nEqPairs < nEqPairs_old) &&
	    (quality < quality_sav) && (total_OpnPen < 10) &&
	    ((RMS > RMSsav) || (RMSsav < BAD_THRESH))) {
	  /* increase gap penalties -
	     ratio of 1/0.05 btwn Opn and Ext == Structal original
	     of 10/0.5 */
	  again = 1;
	  again2 = 1;
	  increased_gap_penalties = 1;
	  printf("Increasing gap penalties\n");
	  total_OpnPen++; /* Should thresh of 10 be adjusted for changes in
			     structal formula constants? XXX */
	  for (j=1; j < (cols-1); j++) {
	    jOpnPen[j]++;
	    jExtPen[j] += 0.05;
	  }
	  for (i=1; i < (rows-1); i++) {
	    iOpnPen[i]++;
	    iExtPen[i] += 0.05;
	  }
	} else if (((quality > quality_sav) && ((RMS < BAD_THRESH) ||
						(RMS <= RMSsav))) ||
		   ((nEqPairs_sav < 0) && ((nEqPairs > nEqPairs_old) ||
					   (quality >= quality_sav) ||
					   ((nEqPairs == nEqPairs_old) &&
					    (RMS <= RMSsav)))) ||
		   ((nbrks < nbrks_sav) && (quality_sav <= quality) &&
		    ((RMS < BAD_THRESH) || (RMS <= RMSsav))) ||
		   ((nEqPairs_sav < nEqPairs) &&
		    ((RMS < BAD_THRESH) || (RMS <= RMSmin)) &&
		    ((RMS - RMSsav) < 0.1))) {
	  if (nEqPairs_sav < 0) {
	    again = 1;
	    if (((quality_sav - quality) >= 0.5) ||
		((quality < quality_sav) &&
		 (nEqPairs < nEqPairs_old))) {
	      printf("Structal method may have failed\n");
	      again2 = 1;
	    }
	  }
	  if (quality > quality_max)
	    quality_max = quality;
	  quality_sav = quality;
	  RMSsav = RMS;
	  nEqPairs_sav = nEqPairs;
	  nbrks_sav = nbrks;
	  sav_updated = 1;
	  if (RMS < RMSmin) {
	    RMSmin = RMS;
	    again = 1;
	  } else if (RMS == RMSmin)
	    again = 1;

	  for (i=0; i<rows; i++) {
	    x_que_sav[1][i] = x_que[1][i]; 
	    x_que_sav[2][i] = x_que[2][i]; 
	    x_que_sav[3][i] = x_que[3][i]; 
	  }
	  for (i=0; i<=MAXSEQ; i++) {
	    iina_sav[i] = iina[i];
	    jinb_sav[i] = jinb[i];
	  }
	} else if ((quality_sav == quality) ||
		   ((RMSmin >= BAD_THRESH) && (RMS < RMSsav)) ||
		   ((((quality_max - quality) < 0.001) ||
		     (((quality_max - quality) < 0.5) &&
		      (total_OpnPen >= 10))) &&
		    ((RMSmin - RMS) >= 0.1))) {
	  again = 1;
	  RMSmin = RMS;
	  RMSsav = RMS;
	  quality_sav = quality;
	  nEqPairs_sav = nEqPairs;
	  nbrks_sav = nbrks;
	  sav_updated = 1;
	  for (i=0; i<rows; i++) {
	    x_que_sav[1][i] = x_que[1][i]; 
	    x_que_sav[2][i] = x_que[2][i]; 
	    x_que_sav[3][i] = x_que[3][i]; 
	  }
	  for (i=0; i<=MAXSEQ; i++) {
	    iina_sav[i] = iina[i];
	    jinb_sav[i] = jinb[i];
	  }
	} else { /* increase gap penalties -
		    ratio of 1/0.05 btwn Opn and Ext == Structal original
		    of 10/0.5 */
	  again = 1;
	  again2 = 1;
	  increased_gap_penalties = 1;
	  printf("Increasing gap penalties\n");
	  total_OpnPen++;
	  for (j=1; j < (cols-1); j++) {
	    jOpnPen[j]++;
	    jExtPen[j] += 0.05;
	  }
	  for (i=1; i < (rows-1); i++) {
	    iOpnPen[i]++;
	    iExtPen[i] += 0.05;
	  }
	}
      }

      if (nEqPairs_sav < 0) {
	if (increased_gap_penalties)
	  stay = 1;
	else {
	  printf("Structal method unsuccessful\n");
	  RMSmin = RMSold;
	  RMSsav = RMSold;
	  quality_sav = quality_old;
	  nEqPairs_sav = nEqPairs_old;
	  nbrks_sav = nbrks_old;
	  for (i=0; i<rows; i++) {
	    x_que_sav[1][i] = x_que_old[1][i]; 
	    x_que_sav[2][i] = x_que_old[2][i]; 
	    x_que_sav[3][i] = x_que_old[3][i]; 
	  }
	  for (i=0; i<=MAXSEQ; i++) {
	    iina_sav[i] = iina_old[i];
	    jinb_sav[i] = jinb_old[i];
	  }
	}
      } else if ((nEqPairs == nEqPairs_old) ||
		 (nEqPairs == nEqPairs_sav)) {
	stay = check_eq(iina,jinb,iina_old,jinb_old,
			MIN(nEqPairs,nEqPairs_old));
	if (stay && (! sav_updated) && (! again2) &&
	    (nEqPairs == nEqPairs_sav) && (iter > 1)) {
	  if (! check_eq(iina,jinb,iina_sav,jinb_sav,
			 MIN(nEqPairs,nEqPairs_sav)))
	    stay = 0;
	}

	if (again2) {
	  if ((! stay) && (iter > 1) && (! increased_gap_penalties))
	    printf("Doing again despite equal pairs (again2 is 1)\n");
	  stay = 1;
	} else if (again) {
	  if ((! stay) && (iter > 1)) {
	    if ((fabs(U[0][0]-1) >= epsilon2) ||
		(fabs(U[0][1]) >= epsilon2) ||
		(fabs(U[0][2]) >= epsilon2) ||
		(fabs(U[1][0]) >= epsilon2) ||
		(fabs(U[1][1]-1) >= epsilon2) ||
		(fabs(U[1][2]) >= epsilon2) ||
		(fabs(U[2][0]) >= epsilon2) ||
		(fabs(U[2][1]) >= epsilon2) ||
		(fabs(U[2][2]-1) >= epsilon2) ||
		(fabs(T[0]) >= epsilon2) ||
		(fabs(T[1]) >= epsilon2) ||
		(fabs(T[2]) >= epsilon2)) {
	      printf("Doing again despite equal pairs\n");
	      stay = 1;
	    }
	  } else
	    stay = 1;
	} else {
	  if (stay && (iter > 1) &&
	      (fabs(U[0][0]-1) < epsilon) &&
	      (fabs(U[0][1]) < epsilon) &&
	      (fabs(U[0][2]) < epsilon) &&
	      (fabs(U[1][0]) < epsilon) &&
	      (fabs(U[1][1]-1) < epsilon) &&
	      (fabs(U[1][2]) < epsilon) &&
	      (fabs(U[2][0]) < epsilon) &&
	      (fabs(U[2][1]) < epsilon) &&
	      (fabs(U[2][2]-1) < epsilon) &&
	      (fabs(T[0]) < epsilon2) &&
	      (fabs(T[1]) < epsilon2) &&
	      (fabs(T[2]) < epsilon2)) {
	    /* prevent getting stuck - Allen */
	    for (i = (stay-1); (i <= nEqPairs) && (i <= nEqPairs_old); i++) {
	      if (isascii(a_que[iina_old[i]-1]) &&
		  isascii(a_que[iina[i]-1]) &&
		  (! islower(a_que[iina_old[i]-1])) &&
		  (! islower(a_que[iina[i]-1])) &&
		  isascii(a_tar[jinb_old[i]-1]) &&
		  isascii(a_tar[jinb[i]-1]) &&
		  (! islower(a_tar[jinb_old[i]-1])) &&
		  (! islower(a_tar[jinb[i]-1])) &&
		  ((iina[i] != iina_old[i]) || (jinb[i] != jinb_old[i]))) {
		printf("Looping; old:  %5d %c%c%c %4d :  %5d  :  %5d %c%c%c %4d\n",
		       iina_old[i], aa_que[iina_old[i]-1][0],
		       aa_que[iina_old[i]-1][1],
		       aa_que[iina_old[i]-1][2], 
		       r_num_que[iina_old[i]-1],
		       i,
		       jinb_old[i], aa_tar[jinb_old[i]-1][0],
		       aa_tar[jinb_old[i]-1][1],
		       aa_tar[jinb_old[i]-1][2],
		       r_num_tar[jinb_old[i]-1]);
		printf("Looping; new:  %5d %c%c%c %4d :  %5d  :  %5d %c%c%c %4d\n",
		       iina[i], aa_que[iina[i]-1][0],
		       aa_que[iina[i]-1][1],
		       aa_que[iina[i]-1][2], r_num_que[iina[i]-1],
		       i,
		       jinb[i], aa_tar[jinb[i]-1][0],
		       aa_tar[jinb[i]-1][1],
		       aa_tar[jinb[i]-1][2],
		       r_num_tar[jinb[i]-1]);
		if (((quality_old - quality) >= 0.5) ||
		    ((RMSold >= BAD_THRESH) && ((RMS - RMSold) >= 0.1)) ||
		    ((RMSold < BAD_THRESH) && (quality_old > quality)) ||
		    ((quality_sav - MAX(quality,quality_old)) >= 0.5) ||
		    ((RMSsav >= BAD_THRESH) &&
		     ((MIN(RMS,RMSold) - RMSsav) >= 0.1)) ||
		    ((RMSsav < BAD_THRESH) &&
		     (quality_sav > MAX(quality,quality_old))))
		  break;
	      }
	    }
	    stay = 0;
	  }
	}
      }

      free((RELEASE)(str1));
      free((RELEASE)(str2));

      for (i=0; i<=MAXSEQ; i++) {
	iina_old[i] = iina[i];
	jinb_old[i] = jinb[i];
      }
      
      free_iVector(iina,0, MAXSEQ);
      free_iVector(jinb,0, MAXSEQ);

      fflush(NULL);
      
    } while(stay > 0 && iter < 100); /* End of iteration cycle */
  } /* if(!lsq) */

  printf("RMSmin = %f\n",RMSmin);
  printf("quality_max = %f\n", quality_max);

  /* Performing back-transformation of BOTH structures
     (QUERY will still be PMI-aligned!): */
  (void)FormRotMatrX(Rtemp,angle_c_tar);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);
  printf("Rotating query: x-angle %.10f; %.1f %.1f %.1f; %.1f %.10f %.10f; %.1f %.10f %.10f\n",
	 angle_c_tar,
	 Rtemp[1][1],Rtemp[1][2],Rtemp[1][3],
	 Rtemp[2][1],Rtemp[2][2],Rtemp[2][3],
	 Rtemp[3][1],Rtemp[3][2],Rtemp[3][3]);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);
  (void)FormRotMatrY(Rtemp,angle_b_tar); 
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);
  printf("Rotating query: y-angle %.10f; %.10f %.1f %.10f; %.1f %.1f %.1f; %.10f %.1f %.10f\n",
	 angle_b_tar,
	 Rtemp[1][1],Rtemp[1][2],Rtemp[1][3],
	 Rtemp[2][1],Rtemp[2][2],Rtemp[2][3],
	 Rtemp[3][1],Rtemp[3][2],Rtemp[3][3]);
  (void)TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);
  (void)TRmoveXYZstruc(Tcm_tar,Runity,x_que,rows);
  printf("Moving query: %.10f %.10f %.10f\n",
	 Tcm_tar[1], Tcm_tar[2], Tcm_tar[3]);
  (void)TRmoveXYZstruc(Tcm_tar,Runity,x_tar,cols);

  /* CHECK COORDINATES: */
/*   stream = fopen("rot_query.dat", "w"); */
/*   if (stream == NULL) */
/*     perror("main: Couldn't open file rot_query.dat"); */
/*   else { */
/*     for (i=0; i<rows; i++) { */
/*       fprintf(stream, "%4s%7d%4s  %c%c%c %1s%4d%12.3f %8.3f %8.3f\n", */
/* 	      "ATOM",i+1,"CA",aa_que[i][0], aa_que[i][1], aa_que[i][2],"L", */
/* 	      i+1,x_que[1][i],  x_que[2][i], x_que[3][i]); */
/*     } */
/*     fclose(stream); */
/*   } */

  if (lsq)
    nEqPairs_sav = nEqPairs;

  printf("Final: RMS, eq.pairs, quality = %f %d %f\n",
	 RMSsav,nEqPairs_sav,quality_sav);

  fflush(NULL);

  if (structal) {  /* STRUCTAL-LIKE OUTPUT */
    stream = fopen(structal_out, "w");
    if (stream == NULL)
      fprintf(stderr, "main: Couldn't open file %s: %s", structal_out,
	      strerror(errno));
    else {
      fprintf(stream, "DUMP rmsnoe   %f (RMS deviation, before elim. bad)\n",
	      RMSsav);
      /*
	EQUIV i seqai : k : j seqaj ja
	EQUIV   105 GLN  105  :    86 :     98 VAL   98
      */  
      fprintf(stream, "#EQUIV WHAT i seqa i : j : m seqa m\n");
      for (i=0; i<nEqPairs_sav; i++) {
	if (a_que[iina_sav[i]-1] == a_tar[jinb_sav[i]-1])
	  aa = '*';
	else
	  aa = ' ';
	if (use_carbonyl && (! isascii(a_que[iina_sav[i]-1]))) {
	  if (islower(toascii(a_que[iina_sav[i]-1]))) {
	    fprintf(stream,
	       "EQUIV C   %5d %c%c%c %c %4d :  %5d  : %c  %5d %c%c%c %4d %c\n",
		    iina_sav[i],aa_que[iina_sav[i]-1][0],
		    aa_que[iina_sav[i]-1][1],
		    aa_que[iina_sav[i]-1][2],
		    _toupper(toascii(a_que[iina_sav[i]-1])),
		    r_num_que[iina_sav[i]-1],
		    i+1,_toupper(toascii(a_tar[jinb_sav[i]-1])),
		    jinb_sav[i],
		    aa_tar[jinb_sav[i]-1][0], aa_tar[jinb_sav[i]-1][1],
		    aa_tar[jinb_sav[i]-1][2], 
		    r_num_tar[jinb_sav[i]-1], aa);
	  } else {
	    fprintf(stream,
	       "EQUIV N   %5d %c%c%c %c %4d :  %5d  : %c  %5d %c%c%c %4d %c\n",
		    iina_sav[i],aa_que[iina_sav[i]-1][0],
		    aa_que[iina_sav[i]-1][1],
		    aa_que[iina_sav[i]-1][2], toascii(a_que[iina_sav[i]-1]),
		    r_num_que[iina_sav[i]-1],
		    i+1, toascii(a_tar[jinb_sav[i]-1]), jinb_sav[i],
		    aa_tar[jinb_sav[i]-1][0], aa_tar[jinb_sav[i]-1][1],
		    aa_tar[jinb_sav[i]-1][2],
		    r_num_tar[jinb_sav[i]-1], aa);
	  }
	} else if (use_carbonyl && islower(a_que[iina_sav[i]-1])) {
	  fprintf(stream,
	       "EQUIV O   %5d %c%c%c %c %4d :  %5d  : %c  %5d %c%c%c %4d %c\n",
		  iina_sav[i],aa_que[iina_sav[i]-1][0],
		  aa_que[iina_sav[i]-1][1],
		  aa_que[iina_sav[i]-1][2], _toupper(a_que[iina_sav[i]-1]),
		  r_num_que[iina_sav[i]-1],
		  i+1, _toupper(a_tar[jinb_sav[i]-1]), jinb_sav[i],
		  aa_tar[jinb_sav[i]-1][0], aa_tar[jinb_sav[i]-1][1],
		  aa_tar[jinb_sav[i]-1][2],
		  r_num_tar[jinb_sav[i]-1], aa);
	} else {
	  fprintf(stream,
	       "EQUIV CA  %5d %c%c%c %c %4d :  %5d  : %c  %5d %c%c%c %4d %c\n",
		  iina_sav[i],aa_que[iina_sav[i]-1][0],
		  aa_que[iina_sav[i]-1][1],
		  aa_que[iina_sav[i]-1][2], a_que[iina_sav[i]-1],
		  r_num_que[iina_sav[i]-1],
		  i+1, a_tar[jinb_sav[i]-1], jinb_sav[i],
		  aa_tar[jinb_sav[i]-1][0], aa_tar[jinb_sav[i]-1][1],
		  aa_tar[jinb_sav[i]-1][2],
		  r_num_tar[jinb_sav[i]-1], aa);
	}
      }
      fprintf(stream, "#EQUIV DONE\n\n");
      fprintf(stream, "IXYZ x_oldk, str. 1 original (PDB standard format)\n");
      for (i=0; i<rows; i++) {
	if (use_carbonyl && (! isascii(a_que[i]))) {
	  if (islower(toascii(a_que[i]))) {
	    fprintf(stream,
		    "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		    i+1," C  ",aa_que[i][0], aa_que[i][1],
		    aa_que[i][2],
		    r_num_que[i], x_que_original[1][i],
		    x_que_original[2][i],
		    x_que_original[3][i]);
	  } else {
	   fprintf(stream,
		   "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		   i+1," N  ",aa_que[i][0], aa_que[i][1],
		   aa_que[i][2],
		   r_num_que[i], x_que_original[1][i],
		   x_que_original[2][i], x_que_original[3][i]);
	 }
       } else if (use_carbonyl && islower(a_que[i])) {
	 fprintf(stream,
		 "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		 i+1," O  ",aa_que[i][0], aa_que[i][1], aa_que[i][2],
		 r_num_que[i],x_que_original[1][i],
		 x_que_original[2][i],
		 x_que_original[3][i]);
       } else {
	 fprintf(stream,
		 "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		 i+1," CA ",aa_que[i][0], aa_que[i][1], aa_que[i][2],
		 r_num_que[i],
		 x_que_original[1][i],
		 x_que_original[2][i],
		 x_que_original[3][i]);
       }
     }
     fprintf(stream, "IXYZ DONE\n\n");
     fprintf(stream, "IXYZ x_old, str. 1 rotated (PDB standard format)\n");
     for (i=0; i<rows; i++) {
       if (use_carbonyl && (! isascii(a_que[i]))) {
	 if (islower(toascii(a_que[i]))) {
	   fprintf(stream,
		   "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		   i+1," C  ",aa_que[i][0], aa_que[i][1],
		   aa_que[i][2],
		   r_num_que[i], x_que[1][i], x_que[2][i],x_que[3][i]);
	 } else {
	   fprintf(stream,
		   "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		   i+1," N  ",aa_que[i][0], aa_que[i][1],
		   aa_que[i][2],
		   r_num_que[i], x_que[1][i], x_que[2][i], x_que[3][i]);
	 }
       } else if (use_carbonyl && islower(a_que[i])) {
	 fprintf(stream,
		 "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		 i+1," O  ",aa_que[i][0], aa_que[i][1], aa_que[i][2],
		 r_num_que[i],x_que[1][i],  x_que[2][i], x_que[3][i]);
       } else {
	 fprintf(stream,
		 "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		 i+1," CA ",aa_que[i][0], aa_que[i][1], aa_que[i][2],
		 r_num_que[i],x_que[1][i],  x_que[2][i], x_que[3][i]);
       }
     }
     fprintf(stream, "IXYZ DONE\n\n");
     fprintf(stream, "IXYZ x_new, str. 2 original (PDB standard format)\n");
     for (i=0; i<cols; i++) {
       if (use_carbonyl && (! isascii(a_tar[i]))) {
	 if (islower(toascii(a_tar[i]))) {
	   fprintf(stream,
		   "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		   i+1," C  ",aa_tar[i][0], aa_tar[i][1],
		   aa_tar[i][2],
		   r_num_tar[i], x_tar[1][i], x_tar[2][i],x_tar[3][i]);
	 } else {
	   fprintf(stream,
		   "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		   i+1," N  ",aa_tar[i][0], aa_tar[i][1],
		   aa_tar[i][2],
		   r_num_tar[i], x_tar[1][i], x_tar[2][i], x_tar[3][i]);
	 }
       } else if (use_carbonyl && islower(a_tar[i])) {
	 fprintf(stream,
		 "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		 i+1," O  ",aa_tar[i][0], aa_tar[i][1], aa_tar[i][2],
		 r_num_tar[i],x_tar[1][i],  x_tar[2][i], x_tar[3][i]);
       } else {
	 fprintf(stream,
		 "ATOM  %5d %4s %c%c%c L%4d    %8.3f%8.3f%8.3f  1.00\n",
		 i+1," CA ",aa_tar[i][0], aa_tar[i][1], aa_tar[i][2],
		 r_num_tar[i],x_tar[1][i],  x_tar[2][i], x_tar[3][i]);
       }
     }
     fprintf(stream, "IXYZ DONE\n\n");
     fclose(stream);
   }
  } /* if(structal) */ else {
    stream = fopen("que.dat", "w");
    if (stream == NULL)
      perror("main: Couldn't open file que.dat");
    else {
      fprintf(stream, "%d\n", rows);
      for (i=0; i<rows; i++) { 
	fprintf(stream, "%8.3f %8.3f %8.3f\n",
		x_que[1][i], x_que[2][i], x_que[3][i]);
      }
      fclose(stream);
    }
  }

  /* FREE ALL ALLOCATED MEMORY: */
  /* Not necessary if returning/exiting from main - Allen */
/*   free_iVector(iina_sav,0, MAXSEQ); */
/*   free_iVector(jinb_sav,0, MAXSEQ); */
  
/*   free_dMatrix(s,0,rows-1,0,cols-1); */
/*  free_dMatrix(SS,0,rows-1,0,cols-1); */
/*  free_dMatrix(TrRow,0,rows-1,0,cols-1); */
/*  free_dMatrix(TrCol,0,rows-1,0,cols-1); */
/*  free_iVector(jOpnPen,0, cols-1); */
/*  free_iVector(jExtPen,0, cols-1); */
/*  free_iVector(iOpnPen,0, rows-1); */
/*  free_iVector(iExtPen,0, rows-1); */

/*  free_dMatrix(x_que,1,3,0,rows-1); */
/*  free_dMatrix(x_que_sav,1,3,0,rows-1); */
/*  free_dMatrix(x_que_original,1,3,0,rows-1); */
/*  free_dMatrix(x_tar,1,3,0,cols-1); */

/*  free_cVector(a_que,0,rows-1); */
/*  free_cVector(a_tar,0,cols-1); */

/*  free_cVector(s1,1,StringMax); */
/*  free_cVector(s2,1,StringMax); */
/*  free_cVector(s3,1,StringMax); */
/*  free_cVector(s4,1,StringMax); */

/*  free_cMatrix(aa_que,0,rows-1,0,2); */
/*  free_cMatrix(aa_tar,0,cols-1,0,2); */

 /* freeing memory from PMI alignment: */
/*  free_dVector(Tcm,1,3); */
/*  free_dVector(Tzero,1,3); */
/*  free_dVector(TMPvec,1,3); */
/*  free_dVector(Tcm_que,1,3); */ /* not needed */
/*  free_dVector(Tcm_tar,1,3); */
/*  free_dMatrix(Rtemp,1,3,1,3); */
/*  free_dMatrix(Runity,1,3,1,3); */

 return 0;
}
